/*------------------------------------------------------------------------------
 Copyright (c) CovertJaguar, 2011-2018

 This work (the API) is licensed under the "MIT" License,
 see LICENSE.md for details.
 -----------------------------------------------------------------------------*/

package mods.railcraft.api.charge;

import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.item.ItemStack;

/**
 * Implemented by items like the Overalls to prevent Charge based damage.
 *
 * Created by CovertJaguar on 10/18/2018 for Railcraft.
 *
 * @author CovertJaguar <http://www.railcraft.info>
 */
public interface IChargeProtectionItem {

    /**
     * Called to determine if the item can currently provide protection against Charge damage.
     *
     * @param owner owner of the item
     * @return true if the item is currently providing protection.
     */
    default boolean isZapProtectionActive(ItemStack stack, EntityLivingBase owner) {
        return true;
    }

    /**
     * Called when charge damage is being done and the target is wearing a Charge protection item.
     * You can block all or just some of the damage to the entity, defaults to all.
     * The damage or other cost done to the item is up to the implementation.
     *
     * @param owner        owner of the item
     * @param attackDamage damage to be done to the owner
     * @return A ZapResult object with the resulting stack and damage prevented.
     */
    default ZepResult zap(ItemStack stack, EntityLivingBase owner, float attackDamage) {
        ItemStack resultStack;
        if (owner.getRNG().nextInt(150) == 0
                && stack.attemptDamageItem(1, owner.getRNG(), owner instanceof EntityPlayerMP ? (EntityPlayerMP) owner : null))
            resultStack = ItemStack.EMPTY;
        else
            resultStack = stack;
        return new ZepResult(resultStack, attackDamage);
    }

    class ZepResult {
        public final ItemStack stack;
        public final float damagePrevented;

        public ZepResult(ItemStack stack, float damagePrevented) {
            this.stack = stack;
            this.damagePrevented = damagePrevented;
        }
    }
}
